# A huge mess...

require 'rational.rb'

# Helps iterating all pairs (n, n+1) in an array
module Pairs
	def each_pair
		prev = self.first
		for i in self[1...self.length]
			yield prev, i
			prev = i
		end
	end
end

def assert(bool)
	raise "assertion failed!" unless bool
end

def sequence_deltas(numseq)
	deltas = []
	numseq.extend Pairs
	numseq.each_pair {|a, b| deltas << (b - a)}
	return deltas
end

def deepcopy_array(a)
	r = Array.new(a.length)
	i = 0
	for e in a
		if e.is_a? Array
			r[i] = deepcopy_array(e)
		else
			r[i] = e
		end
		i += 1
	end
	return r
end

module RecursionRoot
	attr_accessor :recursion_root
end
def recformlambda(root, &block)
	block.extend(RecursionRoot)
	block.recursion_root = root
	return block
end

# Detect simple patterns in numbers
# functype can be :general or :recursive
def detect_num_pattern(numseq, functype = :general)
	assert(numseq.length > 1)
	assert false if functype != :general and functype != :recursive
	numseq.extend Pairs
	
	def detect_constant_sum(numseq, functype)
		delta = numseq[1] - numseq[0]
		numseq.each_pair do |a, b|
			if b - a != delta
				delta = nil
				break
			end
		end
		return nil if delta == nil
		if functype == :general
			return lambda {|n| numseq[0] + n * delta}
		else
			return recformlambda(numseq[0]) {|prev| prev + delta}
		end
	end
	
	def detect_constant_ratio(numseq, functype)
		if numseq[0] != 0
			ratio = Rational(numseq[1], numseq[0])
			numseq.each_pair do |b, a|
				if a == 0 or Rational(a, b) != ratio
					ratio = nil
					break
				end
			end
			return nil if ratio == nil
			if functype == :general
				return lambda {|n| (numseq[0] * (ratio ** n)).to_i}
			else
				return recformlambda(numseq[0]) {|prev| (prev * ratio).to_i}
			end
		end
	end
	
	result = detect_constant_sum(numseq, functype)
	result = detect_constant_ratio(numseq, functype) if result == nil
	return result if result != nil
	
	return nil if functype == :recursive # maybe fix if possible?
	
	# Try to find the above patterns in the deltas of the numbers
	dseq = sequence_deltas(numseq)
	dseq.extend Pairs
	dseq_rule = detect_constant_sum(dseq, functype)
	dseq_rule = detect_constant_ratio(dseq, functype) if dseq_rule == nil
	if dseq_rule != nil
		return lambda do |x|
			myself = lambda do |n|
				return numseq[0] if n == 0
				return myself.call(n - 1) + dseq_rule.call(n - 1)
			end
			myself.call(x)
		end
	end
	
	return nil
end


class Element
	attr_reader :value
	attr_reader :was_char
	def initialize(v)
		if v.is_a? String
			assert(v.length == 1)
			@value = v[0]
			@was_char = true
		else
			@value = v
			@was_char = false
		end
	end
	
	def clone
		if @was_char
			Element.new(@value.chr)
		else
			Element.new(@value)
		end
	end
	
	def ==(rhs)
		return false if not rhs.is_a? Element
		(@value == rhs.value and @was_char == rhs.was_char)
	end
	
	def same_type? (rhs)
		(rhs.is_a? Element and @was_char == rhs.was_char)
	end
	
	def valid?
		if @was_char
			@value >= "A"[0] and @value <= "Z"[0]
		else
			@value >= 0
		end
	end
	
	def pass_through_int(&block)
		v = yield(@value)
		if @was_char
			Element.new(v.chr)
		else
			Element.new(v)
		end
	end
	
	def to_s
		if @was_char
			@value.chr
		else
			@value.to_s
		end
	end
end

class Sequence
	attr_reader :size_func, :lengths
	def initialize(size_func, num_members)
		@size_func = size_func
		@data = Array.new(num_members)
		@lengths = Array.new(num_members)
		@longest_member = 0
		for i in 0...num_members
			l = @size_func.call(i)
			@data[i] = Array.new(l)
			@lengths[i] = l
			if l > @lengths[@longest_member]
				@longest_member = i
			end
		end
		#print stacktrace + "\n"
	end
	
	def clone
		s = Sequence.new(@size_func, @data.length)
		s.members = deepcopy_array(@data)
		return s
	end
	
	def ==(rhs)
		return @data == rhs.members
	end
	
	def longest_member
		@longest_member
	end
	def longest_member_length
		@lengths[@longest_member]
	end
	def last_nonnil
		prev = nil
		for y in @data
			for x in y
				return prev if x == nil
			end
			prev = y
		end
		return prev
	end
	
	def members
		@data
	end
	def members=(d)
		@data = d
	end
	
	def member_length(mi)
		@lengths[mi]
	end
	
	def in_bounds? (member, element)
		return !(member > @data.length or @data[member] == nil or element > @data[member].length)
	end
	def [] (member, element)
		raise "out of bounds" if not in_bounds?(member, element)
		return @data[member][element]
	end
	def []= (member, element, value)
		raise "out of bounds" if not in_bounds?(member, element)
		return @data[member][element] = value
	end
	
	def length
		@data.length
	end
	def num_nonnil_members
		n = 0
		for y in @data
			for x in y
				return n if x == nil
			end
			n += 1
		end
		return n
	end
	def first
		@data.first
	end
	def last
		@data.last
	end
	
	def full?
		nonil = true
		@data.flatten.each {|x| nonil = false if x == nil}
		return nonil
	end
	
	# Write an array of members and check for size
	def write!(members)
		for mi in 0...members.length
			m = members[mi]
			if m.length != @lengths[mi]
				raise "tried to write members with bad length to sequence: " + \
					"got #{m.length}, expected #{@data[mi].length} " + \
					"at index #{mi}\n"
			end
			@data[mi] = m
		end
	end
	
	def compatible?(rule, check_value_match = false)
		ans = true
		rule.walk do |x, y, v|
			throw :terminate if y >= @data.length
			if y >= @data.length or x >= @data[y].length or \
			   (check_value_match and self[y, x] != nil and self[y, x] != v)
				#print "*" + y.to_s + "," + x.to_s + ":" + v.to_s + " ->| " + self[y, x].to_s + "*\n" if not (y >= @data.length or x >= @data[y].length)
				ans = false
				throw :terminate
			end
		end
		return ans
	end
	
	def would_be_changed_by_rule?(rule)
		assert(self.compatible?(rule, false))
		ans = false
		rule.walk do |x, y, v|
			throw :terminate if y >= @data.length
			if self[y, x] != v
				ans = true
				break
			end
		end
		return ans
	end
	
	def apply_rule(rule)
		ret = self.clone
		ret.apply_rule! rule
		ret
	end
	
	def apply_rule!(rule)
		assert(self.compatible?(rule, false))
		rule.walk do |x, y, v|
			throw :terminate if y >= @data.length
			self[y, x] = v
		end
		return true
	end
	
	def to_s
		s = ""
		for m in @data
			for e in m
				if e == nil
					s << "?"
				elsif e.was_char
					s << e.value.chr
				else
					s << e.value.to_s
				end
			end
			s << "\n"
		end
		s
	end
end

class Rule
	attr_reader :chars # Applies to characters instead of numbers?
	attr_reader :xroot, :yroot
	attr_reader :xdelta, :ydelta, :xdelta_rate, :ydelta_rate
	attr_accessor :value_f # Recursion formula
	def initialize(chars, xroot, yroot, xdelta, ydelta, xdelta_rate, ydelta_rate, value_f)
		@chars = chars
		@xroot, @yroot = xroot, yroot
		@xdelta, @ydelta = xdelta, ydelta
		@xdelta_rate, @ydelta_rate = xdelta_rate, ydelta_rate
		@value_f = value_f
	end
	
	def clone
		return Rule.new(@chars, @xroot, @yroot, @xdelta, @ydelta, @xdelta_rate, @ydelta_rate, @value_f)
	end
	
	# Returns 1 if this rule is better, -1 if worse, 0 if as good as.
	def compare_to(r)
		if self.yroot < r.yroot
			return 1
		elsif self.yroot > r.yroot
			return -1
		end
		if self.ydelta < r.ydelta
			return 1
		elsif self.ydelta > r.ydelta
			return -1
		end
		if self.xdelta_rate < r.xdelta_rate
			return 1
		elsif self.xdelta_rate > r.xdelta_rate
			return -1
		end
		if self.xdelta < r.xdelta
			return 1
		elsif self.xdelta > r.xdelta
			return -1
		end
		return 0
	end
	
	def compatible?(rule, max_x, max_y)
		mine = {}
		self.walk do |x, y, v|
			throw :terminate if x < 0 or y < 0 or x >= max_x or y >= max_y
			mine[[x, y]] = v
		end
		rule.walk do |x, y, v|
			throw :terminate if x < 0 or y < 0 or x >= max_x or y >= max_y
			m = mine[[x, y]]
			return false if m != nil and m != v
		end
		return true
	end
	
	# Throw :terminate when the walk should end
	def walk(&block)
		x = @xroot
		y = @yroot
		dx = @xdelta
		dy = @ydelta
		v = @value_f.recursion_root.clone if @value_f != nil
		catch (:terminate) do
			while true
				if @value_f != nil
					yield(x, y, v)
					v = v.pass_through_int &@value_f
				else
					yield(x, y, nil)
				end
				x += dx
				y += dy
				dx += @xdelta_rate
				dy += @ydelta_rate
			end
		end
	end
	
	def nextloc(x, y, depth)
		dx = @xdelta + @xdelta_rate * depth
		dy = @ydelta + @ydelta_rate * depth
		return (x+dx), (y+dy)
	end
	
	def to_s
		"#{@xroot}d#{@xdelta}d#{@xdelta_rate}, " + \
		"#{@yroot}d#{@ydelta}d#{@ydelta_rate}"
	end
end


# All rules in a combination must have the same root
class RuleCombination
	attr_reader :rules
	def initialize(initial_rules = [])
		assert initial_rules.is_a?(Array)
		@rules = []
		for r in initial_rules
			self.add_rule r
		end
	end
	
	def add_rule(r)
		assert(r.value_f != nil)
		if not @rules.empty?
			assert(r.xroot == @rules.first.xroot)
			assert(r.yroot == @rules.first.yroot)
		end
		@rules << r
	end
	
	def del_rule(r)
		@rules.delete r
	end
	
	def num_rules
		@rules.length
	end
	
	def include?(r)
		@rules.include? r
	end
	
	def each(&block)
		for r in @rules
			yield r
		end
	end
	
	# Throw :terminate when walk should end
	def walk(&block)
		return nil if rules.empty?
		
		@visited = []
		walk_impl(rules.first.xroot, \
			rules.first.yroot, \
			0, \
			rules.first.value_f.recursion_root.clone, \
			block)
	end
	
private
	@visited = []
	def walk_impl(x, y, depth, v, block)
		for r in rules
			catch (:terminate) do
				if not @visited.include? [x, y]
					block.call(x, y, v)
					
					@visited << [x, y]
				end
				nx, ny = r.nextloc(x, y, depth)
				nv = v.pass_through_int(&r.value_f)
				walk_impl(nx, ny, depth + 1, nv, block) if not @visited.include? [nx, ny]
			end
		end
	end
end

def detect_value_f(seq, rule)
	# Walk through rule path and collect known elements
	values = []
	prev = nil
	rule.walk do |x, y|
		return nil if y >= seq.length or x >= seq.member_length(y)
		e = seq[y, x]
		throw :terminate if e == nil
		return nil if prev != nil and not prev.same_type? e
		
		values << e.value
		prev = e
	end
	return nil if values.length < 2
	
	numpat = detect_num_pattern(values, :recursive)
	return nil if numpat == nil
	
	if rule.chars
		numpat.recursion_root = Element.new(numpat.recursion_root.chr)
	else
		numpat.recursion_root = Element.new(numpat.recursion_root)
	end
	
	# Test that the rule would produce acceptable values
	testrule = rule.clone
	testrule.value_f = numpat
	testrule.walk do |x, y, v|
		throw :terminate if y >= seq.length or x >= seq.member_length(y)
		return nil if not v.valid?
	end
	return numpat
end

# Returns all individual rules applicable to seq.
def detect_rules(seq)
	assert(seq.length > 1)
	
	coverage = Sequence.new(seq.size_func, seq.num_nonnil_members)
	rules = []

	for yroot in 0...(seq.num_nonnil_members)
	for xroot in 0...(seq.last_nonnil.length)
	for ydelta in 1...(seq.num_nonnil_members)
	for xdelta_rate in 0..(seq.longest_member_length / seq.length)
	#for xdelta in 0...(seq.members[1].length)
	for xdelta in 0...(seq.last_nonnil.length)
		next if not seq.in_bounds?(yroot, xroot)
		next if seq[yroot, xroot] == nil
		rule = Rule.new(seq[yroot, xroot].was_char,
				xroot, yroot, \
				xdelta, ydelta, \
				xdelta_rate, 0, nil)
		rule.value_f = detect_value_f(seq, rule) # Checks compatibility
		if rule.value_f != nil
			next if not coverage.would_be_changed_by_rule? rule
			coverage.apply_rule! rule
			rules << rule
			return rules if coverage.full?
		end
	end
	end
	end
	end
	end
	return rules
end

##### Read input #####
InputFile = "patterns.luk"
OutputFile = "patterns.kir"
members = []
needed_members = -1

begin
	File.open(InputFile, "r") do |f|
		needed_members = (f.readline.strip.split /\s+/)[1].to_i
		while not f.eof?
			m = []
			line = f.readline.strip
			next if line.empty?
			line.scan /[a-zA-Z]|\d+/ do |x|
				x = x.to_i if x =~ /\d+/
				m << Element.new(x)
			end
			members << m
		end
	end
rescue => ex
	print "error reading input:\n"
	print "\t" + (ex.backtrace.join "\n\t")
	exit
end


##### Detect pattern in element length #####
member_lengths = []
for m in members
	member_lengths << m.length
end
member_length_pattern = detect_num_pattern(member_lengths)
raise "error determining member length pattern" if member_length_pattern == nil

(members.length + needed_members).times do |n|
	member_lengths[n] = member_length_pattern.call(n)
end

##### Generate initial sequence datastructure #####
seq = Sequence.new(member_length_pattern, members.length + needed_members)
seq.write!(members)

##### Magic #####
rules = detect_rules(seq)

assert(seq.length > 1)

# Sort best to worst
rules.sort! {|a, b| -a.compare_to(b)}
p rules.length

# Form rule combinations and write until full coverage is achieved
coverage = seq.clone
#coverage = Sequence.new(member_length_pattern, members.length + needed_members)
for r1 in rules
	comb = RuleCombination.new [r1]
	for r2 in rules
		next if r1.xroot != r2.xroot or r1.yroot != r2.yroot
		comb.add_rule r2 if r1 != r2
		print "#{r1}<->#{r2}\n" if r2.xroot == 2 and r1 != r2
		if coverage.compatible?(comb, false) and seq.compatible?(comb, true)
			coverage.apply_rule! comb
			print "-------------------------\n"
			for r in comb
				print r.to_s+"\n"
			end
			print coverage.to_s
		else
			comb.del_rule r2 if r1 != r2
		end
		break if coverage.full?
	end
	break if coverage.full?
end

##### Output #####

File.open(OutputFile, "w") do |f|
	if not coverage.full?
		f.print "Emmosaa!\n"
	else
		for i in (members.length)...(members.length + needed_members)
			for e in coverage.members[i]
				f.print e.to_s
			end
			f.print "\n"
		end
	end
end
