#include <string>
#include <iostream>

#include "patternsolver.h"
#include "strtonumsolver.h"
#include "alphasolver.h"
#include "additionalsolver.h"
#include "repeatingsolver.h"
#include "cutsolver.h"
#include "symmetricaddsolver.h"
#include "splitsolver.h"
#include "trimsolver.h"
#include "slicesolver.h"
#include "simplerepeatingsolver.h"

//-----------------------------------------------------------------------------
PatternSolver::PatternSolver()
//-----------------------------------------------------------------------------
    {
    }

//-----------------------------------------------------------------------------
PatternSolver::~PatternSolver()
//-----------------------------------------------------------------------------
    {
    DeleteSolvers(); // just in case
    }

//-----------------------------------------------------------------------------
void PatternSolver::DeleteSolvers()
//-----------------------------------------------------------------------------
    {
    int count = iSolvers.size();
    for( int i = 0; i < count; i++ )
        {
        delete iSolvers[i];
        }
    iSolvers.clear();
    }

//-----------------------------------------------------------------------------
void PatternSolver::CreateSolvers( int aLevel )
//-----------------------------------------------------------------------------
    {
    // Simple solvers
    AddSolver( new StrToNumSolver(), aLevel );
    AddSolver( new AlphaSolver(), aLevel );
    AddSolver( new AdditionalSolver(), aLevel );
    AddSolver( new CutSolver(), aLevel );
    AddSolver( new SymmetricAddSolver(), aLevel );
    AddSolver( new SimpleRepeatingSolver(), aLevel );

    // Recursive solvers
    AddSolver( new TrimSolver(), aLevel );
    AddSolver( new SplitSolver(), aLevel );
    AddSolver( new RepeatingSolver(), aLevel );
    AddSolver( new SliceSolver(), aLevel );
    }

//-----------------------------------------------------------------------------
void PatternSolver::AddSolver( SolverStringBase* aSolver, int aLevel )
//-----------------------------------------------------------------------------
    {
    aSolver->SetMaxIteration( min( aLevel, iMaxIteration - 1) );
    iSolvers.push_back( aSolver );
    }

//-----------------------------------------------------------------------------
bool PatternSolver::Solve()
//-----------------------------------------------------------------------------
    {
    if( iMaxIteration <= 0 )
        return false;
    if( iPattern.size() < 3 )
        return false;
    iAnswer.clear();

    if( TryLevel( iMaxIteration - 1 ) )
        return true;
    return false;
    }

//-----------------------------------------------------------------------------
bool PatternSolver::IterativeSolve()
//-----------------------------------------------------------------------------
    {
    iAnswer.clear();
    if( iMaxIteration <= 0 )
        return false;
    if( iPattern.size() < 3 )
        return false;

    // Find 'shortest' answer. This does a lot of same work, but it's more
    // accurate.
    for( int i = 2; i <= min( 32, iMaxIteration ); i++ )
        {
        if( TryLevel( i ) )
            {
            return true;
            }
        }
    return false;
    }
//-----------------------------------------------------------------------------
bool PatternSolver::TryLevel( int aLevel )
//-----------------------------------------------------------------------------
    {
    CreateSolvers( aLevel );
    int count = iSolvers.size();
    for( int i = 0; i < count; i++ )
        {
        iSolvers[i]->SetPattern( iPattern, iRequired );
        if( iSolvers[i]->Solve() )
            {
            iSolvers[i]->GetAnswer( iAnswer );
            iSolverStr = iSolvers[i]->SolveString();
            DeleteSolvers();
            return true;
            }
        }
    DeleteSolvers();
    return false;
    }
