#include <sstream>
#include <iostream>
#include "additionalsolver.h"
#include "patternsolver.h"
#include "solverutils.h"

using namespace std;

//-----------------------------------------------------------------------------
bool AdditionalSolver::SolveAdditional( int aAddSize )
//-----------------------------------------------------------------------------
    {
    // Test that every length increases with aAddSize
    int count = iPattern.size();
    int baselength = iPattern[0].length();
    for( int i = 0; i < count; i++ )
        {
        if( iPattern[i].size() != baselength + aAddSize * i )
            return false;
        }

    // Test that previous pattern is the start of the following one
    for( int i = 0; i < count - 1; i++ )
        {
        if( iPattern[i+1].find( iPattern[i], 0 ) != 0 )
            return false;
        }
    
    // Slice last pattern into pieces and solve that
    vector<string> alpha;
    int length = iPattern[count - 1].size();
    for( int i = 0; i < length; i++ )
        {
        string str( iPattern[count - 1].substr( i, 1 ) );
        alpha.push_back( str );
        }

    SolverStringBase* slv = new PatternSolver();
    slv->SetMaxIteration( iMaxIteration - 1 );
    slv->SetPattern( alpha, iRequired * aAddSize );
    if( slv->Solve() )
        {
        // solved it
        alpha.clear();
        slv->GetAnswer( alpha );
        iSolverStr = "(adding)" + slv->SolveString();
        delete slv;

        // construct answer
        for( int i = 0; i < iRequired; i++ )
            {
            string str( iPattern[count - 1] );
            for( int j = 0; j < (i + 1) * aAddSize; j++ )
                {
                str += alpha[j];
                }
            iAnswer.push_back( str );
            }
        return true;
        }
    // could not solve
    delete slv;
    return false;
    }

//-----------------------------------------------------------------------------
// Solve additional patterns e.g.
// A,AB,ABC,ABCD
//
bool AdditionalSolver::Solve()
//-----------------------------------------------------------------------------
    {
    iAnswer.clear();
    if( iPattern.size() < 3 )
        return false;
    if( iMaxIteration <= 0 )
        return false;
        
    // Try different addition sizes
    for( int i = 1; i < 6; i++ )
        {
        if( SolveAdditional( i ) )
            return true;
        }

    // Solve reversed
    SolverUtils::ReverseStringArray( iPattern );
    for( int i = 1; i < 6; i++ )
        {
        if( SolveAdditional( i ) )
            {
            SolverUtils::ReverseStringArray( iAnswer );
            return true;
            }
        }
    return false;
    }
