#include <sstream>
#include <iostream>
#include "cutsolver.h"
#include "numericsolver.h"
#include "solverutils.h"

using namespace std;

//-----------------------------------------------------------------------------
bool CutSolver::SolveCut()
//-----------------------------------------------------------------------------
    {
    iAnswer.clear();
    int count = iPattern.size();
    if( iMaxIteration <= 0 )
        return false;
    if( count < 3 )
        return false;
        
    vector<int> lengths;
    vector<int> positions;
    positions.push_back( 0 );
    
    // Test that each element is part of the previous one
    for( int i = 0; i < count - 1; i++ )
        {
        string::size_type pos = iPattern[0].find( iPattern[i+1], 0 );
        if( pos == string::npos )
            return false;
        positions.push_back( pos );
        lengths.push_back( iPattern[i].length() );
        }
    lengths.push_back( iPattern[count-1].length() );

    NumericSolver* solver = new NumericSolver();
    solver->SetMaxIteration( iMaxIteration - 1 );

    // solve stored positions
    solver->SetPattern( positions, iRequired );
    if( !solver->Solve() )
        {
        // not solvable
        delete solver;
        return false;
        }
    solver->GetAnswer( positions );

    // solve substring lengths
    solver->SetPattern( lengths, iRequired );
    bool solved = false;
    if( solver->Solve() )
        {
        solved = true;
        solver->GetAnswer( lengths );
        for( int i = 0; i < iRequired; i++ )
            {
            if( lengths[i] <= 0 )
                {
                solved = false;
                break;
                }
            string tmp = iPattern[0].substr( positions[i], lengths[i] );
            iAnswer.push_back( tmp );
            }
        }

    iSolverStr = "(cut)" + solver->SolveString();

    delete solver;
    return solved;
    }

//-----------------------------------------------------------------------------
// Solve patterns in which [n+1] element is found inside [n] e.g.
// ABCDE,ABCD,ABC
//
bool CutSolver::Solve()
//-----------------------------------------------------------------------------
    {
    if( SolveCut() )
        return true;

    // Solve reversed
    SolverUtils::ReverseStringArray( iPattern );
    if( SolveCut() )
        {
        SolverUtils::ReverseStringArray( iAnswer );
        return true;
        }

    return false;
    }
