#include <iostream>
#include "numericsolver.h" 
#include "solverutils.h"
using namespace std;

//-----------------------------------------------------------------------------
bool NumericSolver::Solve()
//-----------------------------------------------------------------------------
    {
    iAnswer.clear();
    iSolverStr.clear();
    
    if( iMaxIteration <= 0 )
        return false;
    
    if( iPattern.size() < 3 )
        return false;
    
    // try solving...
    if( SolveAdditional() )
        return true;

    if( SolveMultiply() )
        return true;

    if( SolveDivision() )
        return true;

    if( SolveFibo() )
        return true;

    if( SolveDerivate() ) // find pattern in derivate
        return true;
        
    return false;
    }

//-----------------------------------------------------------------------------
bool NumericSolver::SolveAdditional()
//-----------------------------------------------------------------------------
    {
    int addition = iPattern[1] - iPattern[0];
    int count = iPattern.size();
    for( int i = 1; i < count - 1; i++ )
        {
        if( iPattern[i+1] - iPattern[i] != addition )
            return false;
        }
    
    // write answer
    int start = iPattern[count-1];
    for( int i = 0; i < iRequired; i++ )
        {
        start += addition;
        iAnswer.push_back( start );
        }
    if( addition != 0 )
        {
        iSolverStr = "(add)";
        }
    else
        {
        iSolverStr = "(add:same)"; // special case
        }
    return true;
    }

//-----------------------------------------------------------------------------
bool NumericSolver::SolveMultiply()
//-----------------------------------------------------------------------------
    {
    if( iPattern[0] == 0 || abs(iPattern[1]) < iPattern[0] )
        return false;

    int mul = iPattern[1] / iPattern[0];
    if( mul == 0 )
        return false;

    int count = iPattern.size();
    for( int i = 1; i < count - 1; i++ )
        {
        if( abs(iPattern[i+1]) < iPattern[i] )
            return false;
        if( iPattern[i+1] != iPattern[i] * mul )
            return false;
        }
    
    // write answer
    int start = iPattern[count-1];
    for( int i = 0; i < iRequired; i++ )
        {
        start *= mul;
        iAnswer.push_back( start );
        }
    iSolverStr = "(mul)";
    return true;
    }

//-----------------------------------------------------------------------------
bool NumericSolver::SolveDivision()
//-----------------------------------------------------------------------------
    {
    if( iPattern[1] == 0 || iPattern[0] < iPattern[1] )
        return false;

    int div = iPattern[0] / iPattern[1];
    int count = iPattern.size();
    for( int i = 1; i < count - 1; i++ )
        {
        if( iPattern[i] < iPattern[i+1] )
            return false;
        if( iPattern[i] != iPattern[i+1] * div )
            return false;
        }
    
    // write answer
    int start = iPattern[count-1];
    for( int i = 0; i < iRequired; i++ )
        {
        start /= div;
        iAnswer.push_back( start );
        }
    iSolverStr = "(div)";
    return true;
    }

//-----------------------------------------------------------------------------
bool NumericSolver::SolveFibo()
//-----------------------------------------------------------------------------
    {
    int count = iPattern.size();
    for( int i = 0; i < count - 2; i++ )
        {
        if( iPattern[i] + iPattern[i+1] != iPattern[i+2] )
            return false;
        }
    
    // write answer
    int high = iPattern[count-1];
    int low = iPattern[count-2];
    for( int i = 0; i < iRequired; i++ )
        {
        int tmp = low + high;
        low = high;
        high = tmp;
        iAnswer.push_back( tmp );
        }
    iSolverStr = "(fib)";
    return true;
    }

//-----------------------------------------------------------------------------
bool NumericSolver::SolveDerivate()
//-----------------------------------------------------------------------------
    {
    vector<int> derivate;
    
    // calculate derivate
    int count = iPattern.size();
    for( int i = 0; i < count - 1; i++ )
        {
        int dx = iPattern[i+1] - iPattern[i];
        derivate.push_back( dx );
        }

    // and solve it
    SolverIntBase* solver = new NumericSolver();
    solver->SetMaxIteration( min( 1, iMaxIteration - 1 ) ); // only 1 derivate
    solver->SetPattern( derivate, iRequired );
    if( solver->Solve() )
        {
        vector<int> answ;
        solver->GetAnswer( answ );
        int ans = iPattern[ count - 1 ];
        for( int i = 0; i < iRequired; i++ )
            {
            ans += answ[i];
            iAnswer.push_back( ans );
            }
        iSolverStr = "(der)" + solver->SolveString();
        delete solver;
        return true;
        }

    delete solver;
    return false;
    }
