#include <iostream>
#include <sstream>
#include "repeatingsolver.h"
#include "solverutils.h"
#include "patternsolver.h"
using namespace std;

//-----------------------------------------------------------------------------
// Divide patterns into steps and try to solve those separately
// e.g A,1,B,2,C,3 => A,B,C : 1,2,3
//
bool RepeatingSolver::Solve()
//-----------------------------------------------------------------------------
    {
    iAnswer.clear();
    if( iMaxIteration <= 0 )
        return false;
    
    int count = ( 1 + iPattern.size() ) / 2;
    for( int i = 2; i < count; i++ )
        {
        if( TrySteps(i) )
            {
            return true;
            }
        }
    return false;
    }

//-----------------------------------------------------------------------------
bool RepeatingSolver::TrySteps( int aSteps )
//-----------------------------------------------------------------------------
{
    int count = iPattern.size();
    if( aSteps >= count )
        return false;

    vector< vector<string> > newPatterns;
    
    // divide pattern
    for( int j = 0; j < aSteps; j++ )
        {
        vector<string> tmp;
        for( int i = j; i < count; i += aSteps )
            {
            tmp.push_back( iPattern[i] );
            }
        newPatterns.push_back( tmp );
        }

    // create solvestring
    stringstream str;
    str << "(rep:" << aSteps << ")[";

    // solve new step patterns
    PatternSolver* solver = new PatternSolver();
    solver->SetMaxIteration( iMaxIteration - 1 );
    int newcount = newPatterns.size();
    for( int i = 0; i < newcount; i++ )
        {
        solver->SetPattern( newPatterns[i], iRequired / aSteps + 1 );
        if( !solver->Solve() )
            {
            delete solver;
            return false;
            }
        str << solver->SolveString() << " ";
        solver->GetAnswer( newPatterns[i] );
        }
    str << "]";
    iSolverStr = str.str();

    // construct answer
    int j = -1;
    int shift = count % aSteps;

    for( int i = 0; i < iRequired; i++ )
        {
        if( i % aSteps == 0 )
            j++;
        iAnswer.push_back( newPatterns[ ( shift + i ) % aSteps ][j] );
        }

    return true;
}
