#include <sstream>
#include <iostream>
#include "symmetricaddsolver.h"
#include "additionalsolver.h"
#include "solverutils.h"

using namespace std;

//-----------------------------------------------------------------------------
// Solve patterns where something is added to the boths sides of the the
// previous pattern e.g.
// A, BAB, CBABC, DCBABCD, ...
//
bool SymmetricAddSolver::Solve()
//-----------------------------------------------------------------------------
    {
    iAnswer.clear();
    if( iMaxIteration <= 0 )
        return false;

    int count = iPattern.size();
    int firstlen = iPattern[0].length();
    vector<int> lengths;
    lengths.push_back( 0 ); // no addition to the first element
    
    // Test that strings a symmetric
    for( int i = 1; i < count; i++ )
        {
        int add = iPattern[i].length() - firstlen;
        if( ( add < 2 ) || ( add % 2 ) != 0 )
            return false; // not symmetric
        lengths.push_back( add / 2 );
        }

    // take left part
    int substrlen = firstlen;
    vector<string> leftside;
    for( int i = 0; i < count; i++ )
        {
        substrlen = firstlen + lengths[i];
        leftside.push_back( iPattern[i].substr( 0, substrlen ) );
        }

    // take right part
    substrlen = firstlen;
    vector<string> rightside;
    for( int i = 0; i < count; i++ )
        {
        substrlen = firstlen + lengths[i];
        rightside.push_back( iPattern[i].substr( iPattern[i].length() - substrlen, substrlen ) );
        }

    // solve right part
    AdditionalSolver* solver = new AdditionalSolver();
    solver->SetPattern( rightside, iRequired );
    if( !solver->Solve() )
        {
        delete solver;
        return false;
        }
    iSolverStr = "(sym)[" + solver->SolveString();
    solver->GetAnswer( rightside );
    
    // solve left part
    solver->SetPattern( leftside, iRequired );
    if( !solver->Solve() )
        {
        delete solver;
        return false;
        }
    iSolverStr += "][" + solver->SolveString() + "]";
    solver->GetAnswer( leftside );
    delete solver;

    // try to combine results
    if( CombineResults( leftside, rightside, firstlen ) )
        {
        // Solved it
        return true;
        }

    return false;
    }

//-----------------------------------------------------------------------------
bool SymmetricAddSolver::CombineResults( vector<string>& aLeft,
                                         vector<string>& aRight,
                                         int aCommonChars )
//-----------------------------------------------------------------------------
    {
    if( aLeft.size() != aRight.size() )
        return false;

    int count = aLeft.size();
    for( int i = 0; i < count; i++ )
        {
        // if common characters are not the same, then this is not a good answer
        if( aLeft[i].substr( aLeft[i].length() - aCommonChars, aCommonChars ) !=
            aRight[i].substr( 0, aCommonChars ) )
                return false;

        string str = aLeft[i] + aRight[i].substr( aCommonChars, aRight[i].length() - aCommonChars );
        iAnswer.push_back( str );
        }

    return true;
    }
